@@ -32,9 +32,11 @@ import cgi
 import copy
 from cStringIO import StringIO
 from datetime import datetime, date
+from email.Utils import formatdate
 from gzip import GzipFile
 import os
 import simplejson
+import time
 import zlib
 
 # Import SHA in a way compatible with both Python 2.4 and Python 2.6.
@@ -74,6 +76,7 @@ from zope.traversing.browser import absoluteURL, AbsoluteURL
 from zope.traversing.browser.interfaces import IAbsoluteURL
 
 from lazr.batchnavigator import BatchNavigator
+from lazr.delegates import Passthrough
 from lazr.enum import BaseItem
 from lazr.lifecycle.event import ObjectModifiedEvent
 from lazr.lifecycle.snapshot import Snapshot
@@ -87,7 +90,7 @@ from lazr.restful.interfaces import (
     IUnmarshallingDoesntNeedValue, IWebServiceClientRequest,
     IWebServiceConfiguration, IWebServiceLayer, IWebServiceVersion,
     LAZR_WEBSERVICE_NAME)
-from lazr.restful.utils import get_current_browser_request
+from lazr.restful.utils import get_current_web_service_request
 
 
 # The path to the WADL XML Schema definition.
@@ -158,7 +161,7 @@ class ResourceJSONEncoder(simplejson.JSONEncoder):
                 return tuple(obj)
             if isinstance(underlying_object, dict):
                 return dict(obj)
-        request = get_current_browser_request()
+        request = get_current_web_service_request()
         if queryMultiAdapter((obj, request), IEntry):
             obj = EntryResource(obj, request)
 
@@ -277,8 +280,10 @@ class HTTPResource:
 
         This method has side effects. If the write operation should
         not continue, because the incoming ETag doesn't match the
-        generated ETag, it sets the response code to 412
-        ("Precondition Failed").
+        generated ETag, it sets the response code to 412 ("Precondition
+        Failed"). Whether or not the two ETags 'match' depends on the
+        resource-specific implementation, but by default two ETags
+        only 'match' if they are identical.
 
         If the PUT or PATCH request is being tunneled through POST
         with X-HTTP-Method-Override, the media type of the incoming
@@ -307,7 +312,7 @@ class HTTPResource:
             # This is not a conditional write.
             return media_type
         existing_etag = self.getETag(media_type)
-        if existing_etag in incoming_etags:
+        if self._etagMatchesForWrite(existing_etag, incoming_etags):
             # The conditional write can continue.
             return media_type
         # The resource has changed since the client requested it.
@@ -316,11 +321,18 @@ class HTTPResource:
         self.request.response.setStatus(412)
         return None
 
-    def _getETagCore(self, cache=None):
+    def _etagMatchesForWrite(self, existing_etag, incoming_etags):
+        """Perform an ETag match to see if conditional write is OK.
+
+        By default, we match the full ETag strings against each other.
+        """
+        return existing_etag in incoming_etags
+
+    def _getETagCores(self, cache=None):
         """Calculate the core ETag for a representation.
 
-        :return: a string that will be used to calculate the full
-            ETag. If None is returned, no ETag at all will be
+        :return: a list of strings that will be used to calculate the
+            full ETag. If None is returned, no ETag at all will be
             calculated for the given resource and media type.
         """
         return None
@@ -346,75 +358,35 @@ class HTTPResource:
             # The WADL representation of a resource only changes when
             # the software itself changes. Thus, we don't need any
             # special information for its ETag core.
-            etag_core = ''
+            etag_cores = ['']
         else:
             # For other media types, we calculate the ETag core by
             # delegating to the subclass.
-            etag_core = self._getETagCore(cache)
+            etag_cores = self._getETagCores(cache)
 
-        if etag_core is None:
+        if etag_cores is None:
             return None
 
-        hash_object = sha_constructor()
-        hash_object.update(etag_core)
+        core_hashes = []
+        for core in etag_cores:
+            hash_object = sha_constructor()
+            hash_object.update(core)
+            core_hashes.append(hash_object)
 
         # Append the media type, so that web browsers won't treat
         # different representations of a resource interchangeably.
-        hash_object.update("\0" + media_type)
+        core_hashes[-1].update("\0" + media_type)
 
         # Append the revision number, because the algorithm for
         # generating the representation might itself change across
         # versions.
         revno = getUtility(IWebServiceConfiguration).code_revision
-        hash_object.update("\0" + revno)
+        core_hashes[-1].update("\0" + revno)
 
-        etag = '"%s"' % hash_object.hexdigest()
+        etag = '"%s"' % "-".join([core.hexdigest() for core in core_hashes])
         self.etags_by_media_type[media_type] = etag
         return etag
 
-    def applyTransferEncoding(self, representation):
-        """Apply a requested transfer-encoding to the representation.
-
-        'gzip' and 'deflate' are the only supported transfer-encodings.
-
-        This method has side effects. If an encoding was applied, it
-        sets the "Transfer-Encoding" response header to the name of
-        that encoding.
-
-        If the web service configuration prohibits the use of
-        hop-by-hop headers, this method does nothing.
-        """
-        if not getUtility(IWebServiceConfiguration).set_hop_by_hop_headers:
-            return representation
-
-        if representation == "":
-            # Don't compress an empty representation--that makes it nonempty.
-            return ""
-        requested_encodings = self._parseAcceptStyleHeader(
-            self.request.get('HTTP_TE'))
-        infinity = float("infinity")
-
-        try:
-            gzip_pos = requested_encodings.index('gzip')
-        except ValueError:
-            gzip_pos = infinity
-        try:
-            deflate_pos = requested_encodings.index('deflate')
-        except ValueError:
-            deflate_pos = infinity
-
-        if gzip_pos < deflate_pos:
-            self.request.response.setHeader("Transfer-Encoding", "gzip")
-            gzipped = StringIO()
-            file = GzipFile(mode="w", fileobj=gzipped)
-            file.write(representation)
-            file.close()
-            return gzipped.getvalue()
-        elif deflate_pos != infinity:
-            self.request.response.setHeader("Transfer-Encoding", "deflate")
-            return zlib.compress(representation)
-        return representation
-
     def implementsPOST(self):
         """Returns True if this resource will respond to POST.
 
@@ -476,14 +448,30 @@ class HTTPResource:
         """Extract a list of ETags from a header and parse the list.
 
         RFC2616 allows multiple comma-separated ETags.
+
+        If the compensate_for_mod_compress_etag_modification value is set,
+        ETags like '"foo"-gzip' and '"foo-gzip"' will be transformed into
+        '"foo"'.
         """
         header = self.request.getHeader(header_name)
         if header is None:
             return []
+        utility = getUtility(IWebServiceConfiguration)
+        strip_gzip_from_etag = (
+            utility.compensate_for_mod_compress_etag_modification)
+        etags = []
         # We're kind of cheating, because entity tags can technically
         # have commas in them, but none of our tags contain commas, so
         # this will work.
-        return [etag.strip() for etag in header.split(',')]
+        for etag in header.split(','):
+            etag = etag.strip()
+            if strip_gzip_from_etag:
+                if etag.endswith('-gzip'):
+                    etag = etag[:-5]
+                elif etag.endswith('-gzip"'):
+                    etag = etag[:-6] + '"'
+            etags.append(etag)
+        return etags
 
     def _fieldValueIsObject(self, field):
         """Does the given field expect a data model object as its value?
@@ -843,7 +831,7 @@ class ReadOnlyResource(HTTPResource):
                 allow_string += " POST"
             self.request.response.setStatus(405)
             self.request.response.setHeader("Allow", allow_string)
-        return self.applyTransferEncoding(result)
+        return result
 
 
 class ReadWriteResource(HTTPResource):
@@ -885,7 +873,7 @@ class ReadWriteResource(HTTPResource):
                 allow_string += " DELETE"
             self.request.response.setStatus(405)
             self.request.response.setHeader("Allow", allow_string)
-        return self.applyTransferEncoding(result)
+        return result
 
 
 class EntryHTMLView:
@@ -942,7 +930,7 @@ class EntryManipulatingResource(ReadWriteResource):
         propagated to the client.
         """
         changeset = copy.copy(changeset)
-        validated_changeset = {}
+        validated_changeset = []
         errors = []
 
         # The self link and resource type link aren't part of the
@@ -969,12 +957,7 @@ class EntryManipulatingResource(ReadWriteResource):
 
         # For every field in the schema, see if there's a corresponding
         # field in the changeset.
-        # Get the fields ordered by name so that we always evaluate them in
-        # the same order. This is needed to predict errors when testing.
-        for name, field in getFieldsInOrder(self.entry.schema):
-            if name.startswith('_'):
-                # This field is not part of the web service interface.
-                continue
+        for name, field in get_entry_fields_in_write_order(self.entry):
             field = field.bind(self.entry.context)
             marshaller = getMultiAdapter((field, self.request),
                                          IFieldMarshaller)
@@ -1100,7 +1083,7 @@ class EntryManipulatingResource(ReadWriteResource):
                             error = "Validation error"
                         errors.append("%s: %s" % (repr_name, error))
                         continue
-                validated_changeset[field] = (name, value)
+                validated_changeset.append((field, value))
         # If there are any fields left in the changeset, they're
         # fields that don't correspond to some field in the
         # schema. They're all errors.
@@ -1120,8 +1103,10 @@ class EntryManipulatingResource(ReadWriteResource):
 
         # Store the entry's current URL so we can see if it changes.
         original_url = absoluteURL(self.entry.context, self.request)
-        # Make the changes.
-        for field, (name, value) in validated_changeset.items():
+
+        # Make the changes, in the same order obtained from
+        # get_entry_fields_in_write_order.
+        for field, value in validated_changeset:
             field.set(self.entry, value)
         # The representation has changed, and etags will need to be
         # recalculated.
@@ -1131,7 +1116,7 @@ class EntryManipulatingResource(ReadWriteResource):
         event = ObjectModifiedEvent(
             object=self.entry.context,
             object_before_modification=entry_before_modification,
-            edited_fields=validated_changeset.keys())
+            edited_fields=[field for field, value in validated_changeset])
         notify(event)
 
         # The changeset contained new values for some of this object's
@@ -1200,14 +1185,14 @@ class EntryFieldResource(FieldUnmarshallerMixin, EntryManipulatingResource):
 
     do_PATCH = do_PUT
 
-    def _getETagCore(self, unmarshalled_field_values=None):
+    def _getETagCores(self, unmarshalled_field_values=None):
         """Calculate the ETag for an entry field.
 
         The core of the ETag is the field value itself.
         """
         name, value = self._unmarshallField(
             self.context.name, self.context.field)
-        return str(value)
+        return [str(value)]
 
     def _representation(self, media_type):
         """Create a representation of the field value."""
@@ -1265,23 +1250,58 @@ class EntryResource(CustomOperationResourceMixin,
         super(EntryResource, self).__init__(context, request)
         self.entry = getMultiAdapter((context, request), IEntry)
 
-    def _getETagCore(self, unmarshalled_field_values=None):
+    def _getETagCores(self, unmarshalled_field_values=None):
         """Calculate the ETag for an entry.
 
         :arg unmarshalled_field_values: A dict mapping field names to
         unmarshalled values, obtained during some other operation such
         as the construction of a representation.
         """
-        values = []
+        unwritable_values = []
+        writable_values = []
         for name, field in getFieldsInOrder(self.entry.schema):
-            if self.isModifiableField(field, False):
-                if (unmarshalled_field_values is not None
-                    and unmarshalled_field_values.get(name)):
-                    value = unmarshalled_field_values[name]
-                else:
-                    ignored, value = self._unmarshallField(name, field)
-                values.append(decode_value(value))
-        return "\0".join(values).encode("utf-8")
+            if self.isModifiableField(field, True):
+                # The client can write to this value.
+                bucket = writable_values
+            elif self.isModifiableField(field, False):
+                # The client can't write this value, but it still
+                # might change.
+                bucket = unwritable_values
+            else:
+                # This value can never change, and as such does not
+                # need to be included in the ETag.
+                continue
+            if (unmarshalled_field_values is not None
+                and unmarshalled_field_values.get(name)):
+                value = unmarshalled_field_values[name]
+            else:
+                ignored, value = self._unmarshallField(name, field)
+            bucket.append(decode_value(value))
+
+        unwritable = "\0".join(unwritable_values).encode("utf-8")
+        writable = "\0".join(writable_values).encode("utf-8")
+        return [unwritable, writable]
+
+
+    def _etagMatchesForWrite(self, existing_etag, incoming_etags):
+        """Make sure no other client has modified this resource.
+
+        For a conditional read to succeed, the entire ETag must
+        match. But for a conditional write to succeed, only the second
+        half of the ETag must match. This prevents spurious 412 errors
+        on conditional writes where the only fields that changed are
+        read-only fields that can't possibly cause a conflict.
+        """
+        existing_write_portion = existing_etag.split('-', 1)[-1]
+        for etag in incoming_etags:
+            if '-' in etag:
+                incoming_write_portion = etag.rsplit('-', 1)[-1]
+            else:
+                incoming_write_portion = etag
+            if existing_write_portion == incoming_write_portion:
+                return True
+        return False
+
 
     def toDataForJSON(self):
         """Turn the object into a simple data structure."""
@@ -1381,8 +1401,10 @@ class EntryResource(CustomOperationResourceMixin,
 
         # Make sure the representation includes values for all
         # writable attributes.
-        # Get the fields ordered by name so that we always evaluate them in
-        # the same order. This is needed to predict errors when testing.
+        #
+        # Get the fields ordered by schema order so that we always
+        # evaluate them in the same order. This is needed to predict
+        # errors when testing.
         for name, field in getFieldsInOrder(self.entry.schema):
             if not self.isModifiableField(field, True):
                 continue
@@ -1549,17 +1571,17 @@ class ServiceRootResource(HTTPResource):
 
     @property
     def request(self):
-        """Fetch the current browser request."""
-        return get_current_browser_request()
+        """Fetch the current web service request."""
+        return get_current_web_service_request()
 
-    def _getETagCore(self, cache=None):
+    def _getETagCores(self, cache=None):
         """Calculate an ETag for a representation of this resource.
 
         The service root resource changes only when the software
         itself changes. This information goes into the ETag already,
         so there's no need to provide anything.
         """
-        return ''
+        return ['']
 
     def __call__(self, REQUEST=None):
         """Handle a GET request."""
@@ -1572,12 +1594,38 @@ class ServiceRootResource(HTTPResource):
             REQUEST.response.setStatus(405)
             REQUEST.response.setHeader("Allow", "GET")
             result = ""
-        return self.applyTransferEncoding(result)
+        return result
+
+    def setCachingHeaders(self):
+        "How long should the client cache this service root?"
+        user_agent = self.request.getHeader('User-Agent', '')
+        if user_agent.startswith('Python-httplib2'):
+            # XXX leonardr 20100412
+            # bug=http://code.google.com/p/httplib2/issues/detail?id=97
+            #
+            # A client with a User-Agent of "Python/httplib2" (such as
+            # old versions of lazr.restfulclient) gives inconsistent
+            # results when a resource is served with both ETag and
+            # Cache-Control. We check for that User-Agent and omit the
+            # Cache-Control headers if it makes a request.
+            return
+        config = getUtility(IWebServiceConfiguration)
+        caching_policy = config.caching_policy
+        if self.request.version == config.active_versions[-1]:
+            max_age = caching_policy[-1]
+        else:
+            max_age = caching_policy[0]
+        if max_age > 0:
+            self.request.response.setHeader(
+                'Cache-Control', 'max-age=%d' % max_age)
+            # Also set the Date header so that client-side caches will
+            # have something to work from.
+            self.request.response.setHeader('Date', formatdate(time.time()))
 
     def do_GET(self):
-        """Describe the capabilities of the web service in WADL."""
-
+        """Describe the capabilities of the web service."""
         media_type = self.handleConditionalGET()
+        self.setCachingHeaders()
         if media_type is None:
             # The conditional GET succeeded. Serve nothing.
             return ""
@@ -1793,7 +1841,7 @@ class RESTUtilityBase:
 
     def _service_root_url(self):
         """Return the URL to the service root."""
-        request = get_current_browser_request()
+        request = get_current_web_service_request()
         return absoluteURL(request.publication.getApplication(request),
                            request)
 
@@ -1913,3 +1961,36 @@ class EntryAdapterUtility(RESTUtilityBase):
         """The URL to the description of the object's full representation."""
         return "%s#%s-full" % (
             self._service_root_url(), self.singular_type)
+
+
+def get_entry_fields_in_write_order(entry):
+    """Return the entry's fields in the order they should be written to.
+
+    The ordering is intended to 1) be deterministic for a given schema
+    and 2) minimize the chance of conflicts. Fields that are just
+    fields come before fields (believed to be) controlled by
+    mutators. Within each group, fields are returned in the order they
+    appear in the schema.
+
+    :param entry: An object that provides IEntry.
+    :return: A list of 2-tuples (field name, field object)
+    """
+    non_mutator_fields = []
+    mutator_fields = []
+    field_implementations = entry.__class__.__dict__
+    for name, field in getFieldsInOrder(entry.schema):
+        if name.startswith('_'):
+            # This field is not part of the web service interface.
+            continue
+
+        # If this field is secretly a subclass of lazr.delegates
+        # Passthrough (but not a direct instance of Passthrough), put
+        # it at the end -- it's probably controlled by a mutator.
+        implementation = field_implementations[name]
+        if (issubclass(implementation.__class__, Passthrough)
+            and not implementation.__class__ is Passthrough):
+            mutator_fields.append((name, field))
+        else:
+            non_mutator_fields.append((name, field))
+    return non_mutator_fields + mutator_fields
+
