@@ -60,7 +60,7 @@ from lazr.restful import (
     Collection, Entry, EntryAdapterUtility, ResourceOperation, ObjectLink)
 from lazr.restful.security import protect_schema
 from lazr.restful.utils import (
-    camelcase_to_underscore_separated, get_current_browser_request,
+    camelcase_to_underscore_separated, get_current_web_service_request,
     make_identifier_safe, VersionedDict)
 
 LAZR_WEBSERVICE_EXPORTED = '%s.exported' % LAZR_WEBSERVICE_NS
@@ -73,8 +73,16 @@ OPERATION_TYPES = (
     'destructor', 'factory', 'read_operation', 'write_operation',
     REMOVED_OPERATION_TYPE)
 
-# Marker to specify that a parameter should contain the request user.
-REQUEST_USER = object()
+
+class REQUEST_USER:
+    """Marker class standing in for the user of the current request.
+
+    This is passed in to annotations like @call_with. This is a class
+    rather than an object because it's going to be run through
+    copy.deepcopy, and we want 'is REQUEST_USER' to succeed on the
+    copy.
+    """
+    pass
 
 
 def _check_called_from_interface_def(name):
@@ -553,6 +561,7 @@ class mutator_for(_method_annotator):
                 "A field can only have one mutator method for version %s; "
                 "%s makes two." % (_version_name(version), method.__name__ ))
         mutator_annotations[version] = (method, dict(method_annotations))
+        method_annotations['is_mutator'] = True
 
 
 def _free_parameters(method, annotations):
@@ -1022,7 +1031,7 @@ class CollectionEntrySchema:
     def __get__(self, instance, owner):
         """Look up the entry schema that adapts the model schema."""
         if instance is None or instance.request is None:
-            request = get_current_browser_request()
+            request = get_current_web_service_request()
         else:
             request = instance.request
         request_interface = getUtility(
@@ -1154,11 +1163,7 @@ def generate_operation_adapter(method, version=None):
                              "version '%s'" % (method.__name__, version))
     tag = match[0]
     if version is None:
-        # We need to incorporate the version into a Python class name,
-        # but we won't find out the name of the earliest version until
-        # runtime. Use a generic string that won't conflict with a
-        # real version string.
-        version = "__Earliest"
+        version = getUtility(IWebServiceConfiguration).active_versions[0]
 
     bases = (BaseResourceOperationAdapter, )
     if tag['type'] == 'read_operation':
