@@ -52,7 +52,7 @@ __all__ = [
     'IWebServiceVersion',
     ]
 
-from zope.schema import Bool, Int, List, TextLine
+from zope.schema import Bool, Dict, Int, List, Text, TextLine
 from zope.interface import Attribute, Interface
 # These two should really be imported from zope.interface, but
 # the import fascist complains because they are not in __all__ there.
@@ -63,6 +63,9 @@ from zope.publisher.interfaces.browser import (
     IBrowserRequest, IDefaultBrowserLayer)
 from lazr.batchnavigator.interfaces import InvalidBatchSizeError
 
+# Constants for periods of time
+HOUR = 3600 # seconds
+
 # The namespace prefix for LAZR web service-related tags.
 LAZR_WEBSERVICE_NS = 'lazr.restful'
 
@@ -394,6 +397,27 @@ class IWebServiceConfiguration(Interface):
     These are miscellaneous strings that may differ in different web
     services.
     """
+    caching_policy = List(
+        value_type=Int(),
+        default = [7 * 24 * HOUR, 1 * HOUR],
+        title=u"The web service caching policy.",
+        description = u"""A list of two numbers, each to be used in the
+        'max-age' field of the Cache-Control header. The first number is
+        used when serving the service root for any web service version
+        except the latest one. The second number is used when serving the
+        service root for the latest version (which probably changes more
+        often).""")
+
+    service_description = TextLine(
+        title=u"Service description",
+        description=u"""A human-readable description of the web service.
+
+        The description may contain HTML, but if it does, it must be a
+        valid XHTML fragment.
+        """,
+        default=u"",
+        )
+
     view_permission = TextLine(
         title=u"View permission", default=u"zope.View",
         description=u"The permission to use when checking object visibility.")
@@ -448,6 +472,33 @@ class IWebServiceConfiguration(Interface):
 
         This list must contain at least one version name.""")
 
+    version_descriptions = Dict(
+        key_type = TextLine(),
+        value_type = Text(),
+        title = u"Human-readable descriptions of the web service versions.",
+        description = u"""A dictionary mapping version names to
+        human-readable descriptions. The descriptions should describe
+        what distinguishes this version from other versions, and
+        mention if/when the version will be removed.
+
+        The descriptions may contain HTML, but if they do, they must be
+        valid XHTML fragments.
+        """,
+        default = {}
+        )
+
+    last_version_with_mutator_named_operations = TextLine(
+        default=None,
+        description=u"""In earlier versions of lazr.restful, mutator methods
+        were also published as named operations. This redundant
+        behavior is no longer enabled by default, but this setting
+        allows for backwards compatibility.
+
+        Mutator methods will also be published as named operations in
+        the version you specify here, and in any previous versions. In
+        all subsequent versions, they will not be published as named
+        operations.""")
+
     code_revision = TextLine(
         default=u"",
         description=u"""A string designating the current revision
@@ -475,19 +526,38 @@ class IWebServiceConfiguration(Interface):
         "a collection, they will not be allowed to request more entries "
         "in the batch than this.")
 
-    set_hop_by_hop_headers = Bool(
-        title=u"Whether your application server will allow lazr.restful "
-        "to set hop-by-hop headers.",
-        default=True,
-        description=u"Ordinarily, lazr.restful automatically compresses  "
-        "outgoing representations and sets the Transfer-Encoding "
-        "header accordingly. (Apache's mod_compress does something "
-        "similar, but it sets the Content-Encoding header, which means "
-        "it has to munge the ETag header, which is no good for our purposes.) "
-        "But WSGI application servers prohibit intermediaries from "
-        "setting hop-by-hop headers like Transfer-Encoding. The best you can "
-        "do in that situation is avoid setting those headers and use a "
-        "web server that does compression using Transfer-Encoding.")
+    compensate_for_mod_compress_etag_modification = Bool(
+        title=u"Accept incoming ETags that appear to have been modified "
+               "in transit by Apache's mod_compress.",
+        default=False,
+
+        description=u"""When mod_compress compresses an outgoing
+        representation, it (correctly) modifies the ETag. But when
+        that ETag comes back in on a conditional GET or PUT request,
+        mod_compress does _not_ transparently remove the modification,
+        because it can't be sure which HTTP intermediary was
+        responsible for modifying the ETag on the way out.
+
+        Setting this value to True will tell lazr.restful that
+        mod_compress is in use. lazr.restful knows what mod_compress
+        does to outgoing ETags, and if this value is set, it will
+        strip the modification from incoming ETags that appear to have
+        been modified by mod_compress.
+
+        Specifically: if lazr.restful generates an etag '"foo"',
+        mod_compress will change it to '"foo"-gzip' or '"foo-gzip"',
+        depending on the version. If this value is set to False,
+        lazr.restful will not recognize '"foo"-gzip' or '"foo-gzip"'
+        as being equivalent to '"foo"'. If this value is set to True,
+        lazr.restful will treat all three strings the same way.
+
+        This is a hacky solution, but until Apache issue 39727 is
+        resolved, it's the best way to get the performance
+        improvements of content-encoding in a real setup with multiple
+        layers of HTTP intermediary. (An earlier attempt to use
+        transfer-encoding instead of content-encoding failed because
+        HTTP intermediaries strip the TE header.)
+        """)
 
     def createRequest(body_instream, environ):
         """A factory method that creates a request for the web service.
